#ifndef __LINKED_LIST_H__
#define __LINKED_LIST_H__

/**
 *  A linked list is a common abstract data type.
 *  Each link in the list is a Node.  A Node holds
 *  a value and a pointer to the next Node.  The Linked List
 *  has a pointer to the first Node and supporting operations 
 *  for inserting and removing
 *  elements.
 */

#include <stdlib.h>

/**
 * Our Node class stores some value of type T
 * as well as a pointer to the next Node
 */
template <class T> struct Node {
	T _value;
	Node<T>* _next;
	Node(T value) : _value(value), _next(NULL) { }
};

/**
 * This is somewhat wierd. Notice that I'm declaring
 * the class, but that's all.  Here's why -- below is the function
 *
 * template <class T>
 * bool operator==(const LinkedList<T>& l1, const LinkedList<T>& l2)
 *
 * which needs to know that LinkedList<T> will exist, but must be
 * declared before we declare it a 'friend' in the class.  Normally,
 * we wouldn't have to declare the function twice, but because of the
 * template syntax, when we have friends that are in parameterized classes
 * we have to declare the function as it would appear in the implementation 
 * file and then redclare it as a friend.
 */
template <class T> class LinkedList;

/**
 * This is the first declaration of the operator== function.
 * We redeclare it as a friend function later, inside the class
 */
template <class T>
bool operator==(const LinkedList<T>& l1, const LinkedList<T>& l2);

/**
 * The class definition
 */
template <class T> class LinkedList {
	
	private:
		
		/**
		 * a pointer to the first node in our list
		 */
		Node<T>* _front;
		
		/**
		 * the length of our list
		 */
		long _length;
	
	public:

		/**
		 * default constructor
		 */
		LinkedList();

		/**
		 * copy constructor
		 */
		LinkedList(const LinkedList<T>& l);
		
		/** 
		 * destructor
		 */
		~LinkedList();
		
		/**
		 * Given a list, this function adds the element value
		 * to the front of the list.  So a list (5 6 7) would 
		 * become (4 5 6 7) if we added 4 to it
		 */
		void add(T value);
		
		/**
		 * Removes the first element of the list.  If the list
		 * is empty, it removes nothing.
		 */
		T remove();
		
		/**
		 * Returns the front of the list, but does not 
		 * remove the element.
		 */
		T front() const;
		
		/**
		 * Tells us if our list is empty or not
		 */
		inline bool empty() const;
		
		/**
		 * Tells us the size of our list
		 */
		long size() const;
		
		/**
		 * The assignment operator
		 */
		LinkedList<T>& operator=(const LinkedList<T>& l);
		
		/**
		 * Copies all the elements from 'node' onward
		 * to this object.
		 */
		void copyElements(Node<T>* node);
		
		/**
		 * Deletes all the elements from _front onward
		 */
		void deleteElements();

		//  OIT's g++ is old, so it doesn't like this syntax.
		// If you are compiling on OIT, you must change the declaration to:
		// friend bool operator==(const LinkedList<T>& l1, const LinkedList<T>& l2);
		// thus eliminating the '<>' after the function name.
		// Newer compilers require the '<>' for friend functions of templates
		friend bool operator==<>(const LinkedList<T>& l1, const LinkedList<T>& l2);
};

#endif // __LINKED_LIST_H__